from enum import Enum
from typing import Any, Dict, List, Optional

"""
======================================================================
 Project: Gedcom-X
 File:    Event.py
 Author:  David J. Cartwright
 Purpose: Python Object representation of GedcomX Event Type, EventType, EventRole, EventRoleType Types

 Created: 2025-08-25
 Updated:
   - 2025-08-31: fixed mutible [] in init, replaced List[Identifer] with IdentifierList
   - 2025-09-03: _from_json_ refactor
   - 2025-09-04: changed _as_dict_ to std dict creation and removed call to serailization
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Type Imports
======================================================================
"""
from .attribution import Attribution
from .conclusion import ConfidenceLevel, Conclusion
from .date import Date
from .evidence_reference import EvidenceReference
from .identifier import IdentifierList
from .note import Note
from .place_reference import PlaceReference
from .resource import Resource
from .source_reference import SourceReference
from .subject import Subject
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class EventRoleType(Enum):
    Principal = "http://gedcomx.org/Principal"
    Participant = "http://gedcomx.org/Participant"
    Official = "http://gedcomx.org/Official"
    Witness = "http://gedcomx.org/Witness"
    
    @property
    def description(self):
        descriptions = {
            EventRoleType.Principal: "The person is the principal person of the event. For example, the principal of a birth event is the person that was born.",
            EventRoleType.Participant: "A participant in the event.",
            EventRoleType.Official: "A person officiating the event.",
            EventRoleType.Witness: "A witness of the event."
        }
        return descriptions.get(self, "No description available.")

class EventRole(Conclusion):
    identifier = 'http://gedcomx.org/v1/EventRole'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,
                 id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = [],
                 analysis: Optional[Resource] = None,
                 notes: Optional[List[Note]] = [],
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None,
                 person: Optional[Resource] = None,
                 type: Optional[EventRoleType] = None,
                 details: Optional[str] = None) -> None:
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution)
        self.person = person
        self.type = type
        self.details = details
        
    def __str__(self) -> str:
        parts = []
        if self.type is not None:
            # assume enums expose .name
            parts.append(f"type={getattr(self.type, 'name', str(self.type))}")
        if self.person is not None:
            # assume classes have meaningful __str__
            parts.append(f"person={self.person}")
        if self.details:
            parts.append(f"details={self.details!r}")
        if getattr(self, "id", None):
            parts.append(f"id={self.id!r}")
        return f"EventRole({', '.join(parts)})" if parts else "EventRole()"

    def __repr__(self) -> str:
        # assume enums expose .name and .value
        if self.type is not None:
            tcls = self.type.__class__.__name__
            tname = getattr(self.type, "name", str(self.type))
            tval = getattr(self.type, "value", self.type)
            type_repr = f"<{tcls}.{tname}: {tval!r}>"
        else:
            type_repr = "None"
        return (
            f"{self.__class__.__name__}("
            f"id={getattr(self, 'id', None)!r}, "
            f"lang={getattr(self, 'lang', None)!r}, "
            f"type={type_repr}, "
            f"person={self.person!r}, "
            f"details={self.details!r})"
        )
    
    @property
    def _as_dict_(self):
        type_as_dict = super()._as_dict_
        if self.person:
            type_as_dict['person'] = Resource(target=self.person)._as_dict_
        if self.type is not None:
            type_as_dict['type'] = getattr(self.type, 'value', self.type)
        if self.details:
            type_as_dict['details'] = self.details

        return type_as_dict if type_as_dict != {} else None


    @classmethod
    def _from_json_(cls,data,context):
        
        if not isinstance(data, dict):
            print("Event: from:",data)
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)}")
        event_role_data = Conclusion._dict_from_json_(data,context)
        event_role_data: Dict[str, Any] = {}

        # person: Person | Resource | URI string
        if (p := data.get("person")) is not None:
            event_role_data["person"] = Resource._from_json_(p, context)
            

        # type: EventRoleType (enum-ish)
        if (typ := data.get("type")) is not None:
            event_role_data["type"] = EventRoleType(typ)
            

        # details: TextValue | str
        if (det := data.get("details")) is not None:
            event_role_data["details"] = det  # str or passthrough

        return cls(**event_role_data)
    
class EventType(Enum):
    Adoption = "http://gedcomx.org/Adoption"
    AdultChristening = "http://gedcomx.org/AdultChristening"
    Annulment = "http://gedcomx.org/Annulment"
    Baptism = "http://gedcomx.org/Baptism"
    BarMitzvah = "http://gedcomx.org/BarMitzvah"
    BatMitzvah = "http://gedcomx.org/BatMitzvah"
    Birth = "http://gedcomx.org/Birth"
    Blessing = "http://gedcomx.org/Blessing"
    Burial = "http://gedcomx.org/Burial"
    Census = "http://gedcomx.org/Census"
    Christening = "http://gedcomx.org/Christening"
    Circumcision = "http://gedcomx.org/Circumcision"
    Confirmation = "http://gedcomx.org/Confirmation"
    Cremation = "http://gedcomx.org/Cremation"
    Death = "http://gedcomx.org/Death"
    Divorce = "http://gedcomx.org/Divorce"
    DivorceFiling = "http://gedcomx.org/DivorceFiling"
    Education = "http://gedcomx.org/Education"
    Engagement = "http://gedcomx.org/Engagement"
    Emigration = "http://gedcomx.org/Emigration"
    Excommunication = "http://gedcomx.org/Excommunication"
    FirstCommunion = "http://gedcomx.org/FirstCommunion"
    Funeral = "http://gedcomx.org/Funeral"
    Immigration = "http://gedcomx.org/Immigration"
    LandTransaction = "http://gedcomx.org/LandTransaction"
    Marriage = "http://gedcomx.org/Marriage"
    MilitaryAward = "http://gedcomx.org/MilitaryAward"
    MilitaryDischarge = "http://gedcomx.org/MilitaryDischarge"
    Mission = "http://gedcomx.org/Mission"
    MoveFrom = "http://gedcomx.org/MoveFrom"
    MoveTo = "http://gedcomx.org/MoveTo"
    Naturalization = "http://gedcomx.org/Naturalization"
    Ordination = "http://gedcomx.org/Ordination"
    Retirement = "http://gedcomx.org/Retirement"
    MarriageSettlment = 'https://gedcom.io/terms/v7/MARS'
    
    @property
    def description(self):
        descriptions = {
            EventType.Adoption: "An adoption event.",
            EventType.AdultChristening: "An adult christening event.",
            EventType.Annulment: "An annulment event of a marriage.",
            EventType.Baptism: "A baptism event.",
            EventType.BarMitzvah: "A bar mitzvah event.",
            EventType.BatMitzvah: "A bat mitzvah event.",
            EventType.Birth: "A birth event.",
            EventType.Blessing: "An official blessing event, such as at the hands of a clergy member or at another religious rite.",
            EventType.Burial: "A burial event.",
            EventType.Census: "A census event.",
            EventType.Christening: "A christening event at birth. Note: use AdultChristening for a christening event as an adult.",
            EventType.Circumcision: "A circumcision event.",
            EventType.Confirmation: "A confirmation event (or other rite of initiation) in a church or religion.",
            EventType.Cremation: "A cremation event after death.",
            EventType.Death: "A death event.",
            EventType.Divorce: "A divorce event.",
            EventType.DivorceFiling: "A divorce filing event.",
            EventType.Education: "An education or educational achievement event (e.g., diploma, graduation, scholarship, etc.).",
            EventType.Engagement: "An engagement to be married event.",
            EventType.Emigration: "An emigration event.",
            EventType.Excommunication: "An excommunication event from a church.",
            EventType.FirstCommunion: "A first communion event.",
            EventType.Funeral: "A funeral event.",
            EventType.Immigration: "An immigration event.",
            EventType.LandTransaction: "A land transaction event.",
            EventType.Marriage: "A marriage event.",
            EventType.MilitaryAward: "A military award event.",
            EventType.MilitaryDischarge: "A military discharge event.",
            EventType.Mission: "A mission event.",
            EventType.MoveFrom: "An event of a move (i.e., change of residence) from a location.",
            EventType.MoveTo: "An event of a move (i.e., change of residence) to a location.",
            EventType.Naturalization: "A naturalization event (i.e., acquisition of citizenship and nationality).",
            EventType.Ordination: "An ordination event.",
            EventType.Retirement: "A retirement event."
        }
        return descriptions.get(self, "No description available.")

    @staticmethod
    def guess(description):
        keywords_to_event_type = {
            "adoption": EventType.Adoption,
            "christening": EventType.Christening,
            "annulment": EventType.Annulment,
            "baptism": EventType.Baptism,
            "bar mitzvah": EventType.BarMitzvah,
            "bat mitzvah": EventType.BatMitzvah,
            "birth": EventType.Birth,
            "blessing": EventType.Blessing,
            "burial": EventType.Burial,
            "census": EventType.Census,
            "circumcision": EventType.Circumcision,
            "confirmation": EventType.Confirmation,
            "cremation": EventType.Cremation,
            "death": EventType.Death,
            "divorce": EventType.Divorce,
            "divorce filing": EventType.DivorceFiling,
            "education": EventType.Education,
            "engagement": EventType.Engagement,
            "emigration": EventType.Emigration,
            "excommunication": EventType.Excommunication,
            "first communion": EventType.FirstCommunion,
            "funeral": EventType.Funeral,
            "arrival": EventType.Immigration,
            "immigration": EventType.Immigration,
            "land transaction": EventType.LandTransaction,
            "marriage": EventType.Marriage,
            "military award": EventType.MilitaryAward,
            "military discharge": EventType.MilitaryDischarge,
            "mission": EventType.Mission,
            "move from": EventType.MoveFrom,
            "move to": EventType.MoveTo,
            "naturalization": EventType.Naturalization,
            "ordination": EventType.Ordination,
            "retirement": EventType.Retirement,
        }

        description_lower = description.lower()

        for keyword, event_type in keywords_to_event_type.items():
            if keyword in description_lower:
                return event_type

        return None  # Default to UNKNOWN if no match is found

class Event(Subject):
    identifier = 'http://gedcomx.org/v1/Event'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,
                 id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[Resource] = None,
                 notes: Optional[List[Note]] = None,
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None,
                 extracted: Optional[bool] = False,
                 evidence: Optional[List[EvidenceReference]] = None,
                 media: Optional[List[SourceReference]] = None,
                 #identifiers: Optional[List[Identifier]] = [],
                 identifiers: Optional[IdentifierList] = None,
                 type: Optional[EventType] = None,
                 date: Optional[Date] = None,
                 place: Optional[PlaceReference] = None,
                 roles: Optional[List[EventRole]] = []) -> None:
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution, extracted, evidence, media, identifiers)

        self.type = type if type and isinstance(type, EventType) else None
        self.date = date if date and isinstance(date, Date) else None               
        self.place = place if place and isinstance(place, PlaceReference) else None
        self.roles = roles if roles and isinstance(roles, list) else []
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = {}
        if self.type:
            type_as_dict['type'] = self.type.value
        if self.date:
            type_as_dict['date'] = self.date._as_dict_
        if self.place:
            type_as_dict['place'] = self.place._as_dict_ 
        if self.roles:
            #print(self.roles)
            for role in self.roles:
                if isinstance(role,list):
                    assert False
            type_as_dict['roles'] = [role._as_dict_ for role in self.roles]
        
        return type_as_dict if type_as_dict != {} else None
  
    
    @classmethod
    def _from_json_(cls, data: Dict[str, Any], context: Any = None) -> "Event":
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict, got {type(data)}")

        event_data: Dict[str, Any] = Subject._dict_from_json_(data,context)

        # Enum / type
        if (typ := data.get("type")) is not None:
            event_data["type"] = EventType(typ)
            

        # Objects
        if (dt := data.get("date")) is not None:
            event_data["date"] = Date._from_json_(dt, context)

        if (pl := data.get("place")) is not None:
            event_data["place"] = PlaceReference._from_json_(pl, context)

        # Lists
        if (rls := data.get("roles")) is not None:
            event_data["roles"] = [EventRole._from_json_(r, context) for r in rls]

        return cls(**event_data)


from typing import Any, Optional, Dict
from datetime import datetime, timezone
from dateutil import parser
"""
======================================================================
 Project: Gedcom-X
 File:    date.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json refactored
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class DateFormat:
    def __init__(self) -> None:
        pass

class DateNormalization():
    pass        

class Date:
    identifier = 'http://gedcomx.org/v1/Date'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, original: Optional[str],normalized: Optional[DateNormalization] = None ,formal: Optional[str | DateFormat] = None) -> None:
        self.original = original
        self.formal = formal

        self.normalized: DateNormalization | None = normalized if normalized else None
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = {}
        if self.original:
            type_as_dict['original'] = self.original
        if self.formal:
            type_as_dict['formal'] = self.formal
        return type_as_dict if type_as_dict != {} else None
        return Serialization.serialize_dict(type_as_dict)

    @classmethod
    def _from_json_(cls,data: Any, context=None):
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)} data:{data}")

        date_data: Dict[str, Any] = {}

        # Scalars
        if (orig := data.get("original")) is not None:
            date_data["original"] = orig
        if (formal := data.get("formal")) is not None:
            date_data["formal"] = formal
        
        return cls(**date_data)
        


def date_to_timestamp(date_str: str, assume_utc_if_naive: bool = True, print_definition: bool = True):
    """
    Convert a date string of various formats into a Unix timestamp.

    A "timestamp" refers to an instance of time, including values for year, 
    month, date, hour, minute, second, and timezone.
    """
    # Handle year ranges like "1894-1912" → pick first year
    if "-" in date_str and date_str.count("-") == 1 and all(part.isdigit() for part in date_str.split("-")):
        date_str = date_str.split("-")[0].strip()

    # Parse date
    dt = parser.parse(date_str)

    # Ensure timezone awareness
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc if assume_utc_if_naive else datetime.now().astimezone().tzinfo)

    # Normalize to UTC and compute timestamp
    dt_utc = dt.astimezone(timezone.utc)
    epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
    ts = (dt_utc - epoch).total_seconds()

    # Create ISO 8601 string with full date/time/timezone
    full_timestamp_str = dt_utc.replace(microsecond=0).isoformat()

    
    return ts, full_timestamp_str


DEBUG = False
import logging
import mimetypes
import re
import xml.etree.ElementTree as ET

from typing import Any, Mapping
import math
import shutil

"""
======================================================================
 Project: Gedcom-X
 File:    converter.py
 Author:  David J. Cartwright
 Purpose: convert gedcom versions

 Created: 2025-08-25
 Updated:
   - 2025-09-01: added docs and fixed imports for lowercase module names
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""

from .address import Address
from .agent import Agent
from .attribution import Attribution
from .conclusion import Conclusion
from .coverage import Coverage
from .date import Date, date_to_timestamp
from .document import Document
from .evidence_reference import EvidenceReference
from .exceptions import TagConversionError
from .event import Event, EventType, EventRole, EventRoleType
from .fact import Fact, FactType, FactQualifier
from .gedcom import Gedcom
from .gedcom5x import Gedcom5x, Gedcom5xRecord
from .gedcomx import GedcomX
from .gender import Gender, GenderType
from .group import Group
from .identifier import Identifier, IdentifierType, make_uid, IdentifierList
from .logging_hub import hub, ChannelConfig
from .name import Name, NameType, NameForm, NamePart, NamePartType, NamePartQualifier
from .note import Note
from .online_account import OnlineAccount
from .person import Person
from .place_description import PlaceDescription
from .place_reference import PlaceReference
from .qualifier import Qualifier
from .relationship import Relationship, RelationshipType
from .resource import Resource
from .source_citation import SourceCitation
from .source_description import SourceDescription, ResourceType
from .source_reference import SourceReference, KnownSourceReference
# from .subject import Subject
from .textvalue import TextValue
#from .topleveltypecollection import TopLevelTypeCollection
from .uri import URI
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
convert_log = "gedcomx.convert.GEDCOM5x"
#=====================================================================

hub.start_channel(
    ChannelConfig(
        name=convert_log,
        path=f"logs/{convert_log}.log",
        level=logging.DEBUG,
        rotation="size:10MB:3",   # rotate by size, keep 3 backups
    ))


class GedcomConverter():
    def __init__(self) -> None:
        self.gedcomx = GedcomX()
        self.object_map: dict[Any, Any] = {-1:self.gedcomx}
        self.missing_handler_count = {}
    
    type_name_type = {
        'aka': NameType.AlsoKnownAs
    }
    gedcom_even_to_fact = {
    # Person Fact Types
    "ADOP": FactType.Adoption,
    "CHR": FactType.AdultChristening,
    "EVEN": FactType.Amnesty,  # and other FactTypes with no direct GEDCOM tag
    "BAPM": FactType.Baptism,
    "BARM": FactType.BarMitzvah,
    "BASM": FactType.BatMitzvah,
    "BIRT": FactType.Birth,
    "BIRT, CHR": FactType.Birth,
    "BLES": FactType.Blessing,
    "BURI": FactType.Burial,
    "CAST": FactType.Caste,
    "CENS": FactType.Census,
    "CIRC": FactType.Circumcision,
    "CONF": FactType.Confirmation,
    "CREM": FactType.Cremation,
    "DEAT": FactType.Death,
    "EDUC": FactType.Education,
    "EMIG": FactType.Emigration,
    "FCOM": FactType.FirstCommunion,
    "GRAD": FactType.Graduation,
    "IMMI": FactType.Immigration,
    "MIL": FactType.MilitaryService,
    "NATI": FactType.Nationality,
    "NATU": FactType.Naturalization,
    "OCCU": FactType.Occupation,
    "ORDN": FactType.Ordination,
    "DSCR": FactType.PhysicalDescription,
    "PROB": FactType.Probate,
    "PROP": FactType.Property,
    "RELI": FactType.Religion,
    "RESI": FactType.Residence,
    "WILL": FactType.Will,

    # Couple Relationship Fact Types
    "ANUL": FactType.Annulment,
    "DIV": FactType.Divorce,
    "DIVF": FactType.DivorceFiling,
    "ENGA": FactType.Engagement,
    "MARR": FactType.Marriage,
    "MARB": FactType.MarriageBanns,
    "MARC": FactType.MarriageContract,
    "MARL": FactType.MarriageLicense,
    "SEPA": FactType.Separation,

    # Parent-Child Relationship Fact Types
    # (Note: Only ADOPTION has a direct GEDCOM tag, others are under "EVEN")
    "ADOP": FactType.AdoptiveParent
}
    
    gedcom_even_to_evnt = {
    # Person Fact Types
    "ADOP": EventType.Adoption,
    "CHR": EventType.AdultChristening,
    "BAPM": EventType.Baptism,
    "BARM": EventType.BarMitzvah,
    "BASM": EventType.BatMitzvah,
    "BIRT": EventType.Birth,
    "BIRT, CHR": EventType.Birth,
    "BLES": EventType.Blessing,
    "BURI": EventType.Burial,
    
    "CENS": EventType.Census,
    "CIRC": EventType.Circumcision,
    "CONF": EventType.Confirmation,
    "CREM": EventType.Cremation,
    "DEAT": EventType.Death,
    "EDUC": EventType.Education,
    "EMIG": EventType.Emigration,
    "FCOM": EventType.FirstCommunion,
    
    "IMMI": EventType.Immigration,
    
    "NATU": EventType.Naturalization,
    
    "ORDN": EventType.Ordination,
    

    # Couple Relationship Fact Types
    "ANUL": EventType.Annulment,
    "DIV": EventType.Divorce,
    "DIVF": EventType.DivorceFiling,
    "ENGA": EventType.Engagement,
    "MARR": EventType.Marriage
    
}
    
    @property
    def ignored_tags(self):
        return self.missing_handler_count if self.missing_handler_count != {} else None

    def clean_str(self, text: str | None) -> str:
        # Regular expression to match HTML/XML tags
        if text is None or text.strip() == '':
            return ""
        clean_text = re.sub(r'<[^>]+>', '', text)
        
        return clean_text
    
    def parse_gedcom5x_record(self,record: Gedcom5xRecord):
        if record is not None: 
            with hub.use(convert_log):       
                handler_name = 'handle_' + record.tag.lower()
                
                if hasattr(self,handler_name): 
                    log.info(f'Using {handler_name} to pars Record: {record.describe()}')         
                    handler = getattr(self,handler_name)            
                    handler(record)
                    if record.tag != 'FAM':
                        for sub_record in record.subRecords():
                            log.debug(sub_record.describe())
                            self.parse_gedcom5x_record(sub_record)
                else:
                    if record.tag in self.missing_handler_count:
                        self.missing_handler_count[record.tag] += 1
                    else:
                        self.missing_handler_count[record.tag] = 1
                    
                    log.error(f'Failed Parsing Record: {record.describe()}')
                log.debug(f"{record.tag} with id: {record.xref} has {len(record.subRecords())} subRecords")
                
        else:
            assert False
    
    def handle__apid(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceReference):
            self.object_map[record.level-1].description.add_identifier(Identifier(type=IdentifierType.Other, value=[URI.from_url('APID://' + record.value if record.value else '')]))
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            self.object_map[record.level-1].add_identifier(Identifier(type=IdentifierType.Other,value=[URI.from_url('APID://' + record.value if record.value else '')]))
        else:
            raise ValueError(f"Could not handle '_APID' tag in record {record.describe()}, last stack object {type(self.object_map[record.level-1])}")

    def handle__meta(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            gxobject = Note(text=self.clean_str(record.value if record.value else 'Warning: This NOTE had not content.'))
            self.object_map[record.level-1].add_note(gxobject)
            
            self.object_map[record.level] = gxobject
        else:
            raise ValueError(f"Could not handle 'WWW' tag in record {record.describe()}, last stack object {self.object_map[record.level-1]}")

    def handle__wlnk(self, record: Gedcom5xRecord):
        return self.handle_sour(record)

    def handle_addr(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            # TODO CHeck if URL?
            if record.value is not None and self.clean_str(record.value):
                gxobject = Address(value=self.clean_str(record.value))
            else:
                gxobject = Address()
            self.object_map[record.level-1].address = gxobject
            
            self.object_map[record.level] = gxobject
        else:
            raise ValueError(f"I do not know how to handle an 'ADDR' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_adr1(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].street = self.clean_str(record.value)        
        else:
            raise ValueError(f"I do not know how to handle an 'ADR1' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_adr2(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].street2 = self.clean_str(record.value)        
        else:
            raise ValueError(f"I do not know how to handle an 'ADR2' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_adr3(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].street3 = self.clean_str(record.value)        
        else:
            raise ValueError(f"I do not know how to handle an 'ADR3' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_adr4(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].street4 = self.clean_str(record.value)        
        else:
            raise ValueError(f"I do not know how to handle an 'ADR4' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_adr5(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].street5 = self.clean_str(record.value)        
        else:
            raise ValueError(f"I do not know how to handle an 'ADR5' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_adr6(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].street5 = self.clean_str(record.value)        
        else:
            raise ValueError(f"I do not know how to handle an 'ADR6' tag for a {type(self.object_map[record.level-1])}")
        
    def handle_phon(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].phones.append(self.clean_str(record.value))        
        else:
            raise ValueError(f"I do not know how to handle an '{record.tag}' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_email(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].emails.append(self.clean_str(record.value))        
        else:
            raise ValueError(f"I do not know how to handle an '{record.tag}' tag for a {type(self.object_map[record.level-1])}")
    
    def handle_fax(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            if record.value is not None and self.clean_str(record.value):
                self.object_map[record.level-1].emails.append('FAX:' + (self.clean_str(record.value) if record.value is not None else ''))        
        else:
            raise ValueError(f"I do not know how to handle an '{record.tag}' tag for a {type(self.object_map[record.level-1])}")

    def handle_adop(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Adoption)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            return #TODO
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_auth(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            if record.value is not None and self.gedcomx.agents.byName(record.value):
                gxobject = self.gedcomx.agents.byName(record.value)[0]
            else:
                gxobject = Agent(names=[TextValue(record.value)])
                self.gedcomx.add_agent(gxobject)
            
            self.object_map[record.level-1].author = gxobject
            
            self.object_map[record.level] = gxobject
        else:
            
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_bapm(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Baptism)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_birt(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Birth)
            self.object_map[record.level-1].add_fact(gxobject)
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_buri(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Burial)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_caln(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceReference):
            self.object_map[record.level-1].description.add_identifier(Identifier(type=IdentifierType.Other,value=[URI.from_url('Call Number:' + record.value if record.value else '')]))
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            self.object_map[record.level-1].add_identifier(Identifier(type=IdentifierType.Other,value=[URI.from_url('Call Number:' + record.value if record.value else '')]))
        elif isinstance(self.object_map[record.level-1], Agent):
            pass
            # TODO Why is GEDCOM so shitty? A callnumber for a repository?
        else:
            raise ValueError(f"Could not handle 'CALN' tag in record {record.describe()}, last stack object {type(self.object_map[record.level-1])}")

    def handle_chan(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            date = record.subRecord('DATE')
            if  date is not None:
                self.object_map[record.level-1].created = Date(date[0].value)
        elif isinstance(self.object_map[record.level-1], Agent):
            if self.object_map[record.level-1].attribution is None:
                gxobject = Attribution()
                self.object_map[record.level-1].attribution = gxobject
                self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Person):
            if self.object_map[record.level-1].attribution is None:
                gxobject = Attribution()
                self.object_map[record.level-1].attribution = gxobject
                self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_chr(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Christening)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_city(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None:
                self.object_map[record.level-1].city = self.clean_str(record.value)
            else: raise ValueError('Record had no value')
        else:
            raise ValueError(f"I do not know how to handle an 'CITY' tag for a {type(self.object_map[record.level-1])}")
        
    def handle_conc(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Note):
            gxobject = self.clean_str(str(record.value))
            self.object_map[record.level-1].append(gxobject)
        elif isinstance(self.object_map[record.level-1], Agent):
            gxobject = str(record.value)
            self.object_map[record.level-1]._append_to_name(gxobject)
        elif isinstance(self.object_map[record.level-1], Qualifier):
            gxobject = str(record.value)
            self.object_map[record.level-2].append(gxobject)
        elif isinstance(self.object_map[record.level-1], TextValue):
            #gxobject = TextValue(value=self.clean_str(record.value))
            self.object_map[record.level-1]._append_to_value(record.value)
        elif isinstance(self.object_map[record.level-1], SourceReference):
            self.object_map[record.level-1].append(record.value)
        elif isinstance(self.object_map[record.level-1], Fact):
            self.object_map[record.level-1].notes[0].text += record.value
            
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_cont(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Note):
            gxobject = str("\n" + record.value if record.value else '')
            self.object_map[record.level-1].append(gxobject)
        elif isinstance(self.object_map[record.level-1], Agent):
            gxobject = str("\n" + record.value if record.value else '')
        elif isinstance(self.object_map[record.level-1], Qualifier):
            gxobject = str("\n" + record.value if record.value else '')
            self.object_map[record.level-1].append(gxobject)
        elif isinstance(self.object_map[record.level-1], TextValue):
            #gxobject = TextValue(value="\n" + record.value)
            self.object_map[record.level-1]._append_to_value(record.value if record.value else '\n')
        elif isinstance(self.object_map[record.level-1], SourceReference):
            self.object_map[record.level-1].append(record.value)
        elif isinstance(self.object_map[record.level-1], Address):
            self.object_map[record.level-1]._append(record.value)
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)
    
    def handle_crea(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            date = record.subRecord('DATE')
            if  date is not None and date != []:
                self.object_map[record.level-1].created = Date(original=date[0].value)
            else: raise ValueError('DATE had not value')                     
            
        elif isinstance(self.object_map[record.level-1], Agent):
            if self.object_map[record.level-1].attribution is None:
                gxobject = Attribution()
                self.object_map[record.level-1].attribution = gxobject
                
                self.object_map[record.level] = gxobject
            else:
                log.info(f"[{record.tag}] Attribution already exists for SourceDescription with id: {self.object_map[record.level-1].id}")
        else:
            raise ValueError(f"Could not handle '{record.tag}' tag in record {record.describe()}, last stack object {self.object_map[record.level-1]}")
    
    def handle_ctry(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            if record.value is not None:
                self.object_map[record.level-1].country = self.clean_str(record.value)
            else:
                raise ValueError('Recrod had no value')
        else:
            raise ValueError(f"I do not know how to handle an '{record.tag}' tag for a {type(self.object_map[record.level-1])}")
     
    def handle_data(self, record: Gedcom5xRecord) -> None:
        if record.value != '' and record.value == 'None':
            assert False
        self.object_map[record.level] = self.object_map[record.level-1]

    def handle_date(self, record: Gedcom5xRecord):
        if record.parent is not None and record.parent.tag == 'PUBL':
            #gxobject = Date(original=record.value) #TODO Make a parser for solid timestamps
            #self.object_map[0].published = gxobject
            #self.object_map[0].published = date_to_timestamp(record.value) if record.value else None 
            self.object_map[0].published = record.value    
            #
            #self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Event):
            self.object_map[record.level-1].date = Date(original=record.value)
        elif isinstance(self.object_map[record.level-1], Fact):
            self.object_map[record.level-1].date = Date(original=record.value)
        elif record.parent is not None and record.parent.tag == 'DATA' and isinstance(self.object_map[record.level-2], SourceReference):
            gxobject = Note(text='Date: ' + record.value if record.value else '')
            self.object_map[record.level-2].description.add_note(gxobject)
            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            
            self.object_map[record.level-1].ctreated = record.value #TODO String to timestamp
        elif isinstance(self.object_map[record.level-1], Attribution):
            if record.parent is not None and record.parent.tag == 'CREA':
                self.object_map[record.level-1].created = record.value #TODO G7
            elif record.parent is not None and record.parent.tag == "CHAN":
                self.object_map[record.level-1].modified = record.value #TODO G7
        elif record.parent is not None and record.parent.tag in ['CREA','CHAN']:
            pass

        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_deat(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Death)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_even(self, record: Gedcom5xRecord):
        # TODO If events in a @S, check if only 1 person matches?
        if record.value and (not record.value.strip() == ''):
            values = [value.strip() for value in record.value.split(",")]
            for value in values:
                if value in self.gedcom_even_to_fact.keys():
                    if isinstance(self.object_map[record.level-1], Person):
                        gxobject = Fact(type=self.gedcom_even_to_fact[value])
                        self.object_map[record.level-1].add_fact(gxobject)

                        
                        self.object_map[record.level] = gxobject

                    elif isinstance(self.object_map[record.level-1], SourceDescription):
                        gxobject = Event(type=self.gedcom_even_to_evnt[value],sources=[self.object_map[record.level-1]])
                        self.gedcomx.add_event(gxobject)
                        
                        self.object_map[record.level] = gxobject
                    else:
                        log.warning(f"Could not convert EVEN '{value}' for object of type {type(self.object_map[record.level-1])} in record {record.describe()}")
                        return
                        raise TagConversionError(record=record,levelstack=self.object_map)
                        assert False
                        # TODO: Fix, this. making an event to cacth subtags, why are these fact tied to a source? GEDCOM is horrible
                        gxobject = Event(type=EventType.UNKNOWN)
                        
                        self.object_map[record.level] = gxobject
                else:
                    log.warning(f"EVEN type is not known {record.describe()}")
                    gxobject = Event(roles=[EventRole(person=self.object_map[record.level],type=EventRoleType.Principal)])
                    self.gedcomx.add_event(gxobject)
                    self.object_map[record.level] = gxobject
                    

        else:
            possible_fact = FactType.guess(record.subRecord('TYPE')[0].value)
            if possible_fact:
                gxobject = Fact(type=possible_fact)
                self.object_map[record.level-1].add_fact(gxobject)

                
                self.object_map[record.level] = gxobject
                return
            elif EventType.guess(record.subRecord('TYPE')[0].value):
                if isinstance(self.object_map[record.level-1], Person):
                    gxobject = Event(type=EventType.guess(record.subRecord('TYPE')[0].value), roles=[EventRole(person=self.object_map[record.level-1], type=EventRoleType.Principal)])
                    self.gedcomx.add_event(gxobject)
                    
                    self.object_map[record.level] = gxobject
                return
            else:
                if isinstance(self.object_map[record.level-1], Person):
                    gxobject = Event(type=None, roles=[EventRole(person=self.object_map[record.level-1], type=EventRoleType.Principal)])
                    gxobject.add_note(Note(subject='Event', text=record.value))
                    self.gedcomx.add_event(gxobject)
                    
                    self.object_map[record.level] = gxobject
                    return
                    
                else:
                    assert False

    def handle_exid(self,record: Gedcom5xRecord):
        if record.value:
            gxobject = Identifier(type=IdentifierType.External,value=[URI._from_json_(record.value) if record.value else URI()]) # type: ignore
            self.object_map[record.level-1].add_identifier(gxobject)       
            self.object_map[record.level] = gxobject
        else: raise ValueError('Record had no value')

    def handle_fam(self, record: Gedcom5xRecord) -> None:
        if record.tag != 'FAM' or record.level != 0:
            raise ValueError("Invalid record: Must be a level 0 FAM record")

        husband, wife, children = None, None, []

        husband_record = record.subRecords('HUSB')
        if husband_record is not None:
            id = husband_record[0].xref if len(husband_record) > 0 else None
            if id:
                husband = self.gedcomx.get_person_by_id(id)

        wife_record = record.subRecords('WIFE')
        if wife_record:
            id = wife_record[0].xref if len(wife_record) > 0 else None
            if id:
                wife = self.gedcomx.get_person_by_id(id)

        children_records = record.subRecords('CHIL')
        if children_records:
            for child_record in children_records:
                id = child_record.xref
                if id:
                    child = self.gedcomx.get_person_by_id(id)
                    if child:
                        children.append(child)

        if husband:
            for child in children:
                relationship = Relationship(person1=husband, person2=child, type=RelationshipType.ParentChild)
                self.gedcomx.add_relationship(relationship)
        if wife:
            for child in children:
                relationship = Relationship(person1=wife, person2=child, type=RelationshipType.ParentChild)
                self.gedcomx.add_relationship(relationship)
        if husband and wife:
            relationship = Relationship(person1=husband, person2=wife, type=RelationshipType.Couple)
            self.gedcomx.add_relationship(relationship)
            self.object_map[record.level] = relationship
        
        if (marr_record := record.subRecord('MARR')) is not None:
            self.handle_marr(marr_record[0])
    
    def handle_famc(self, record: Gedcom5xRecord) -> None:
        return

    def handle_fams(self, record: Gedcom5xRecord) -> None:
        return

    def handle_file(self, record: Gedcom5xRecord):
        if record.value and record.value.strip() != '':
            #raise ValueError(f"I did not expect the 'FILE' tag to have a value: {record.value}")
            #TODO Handle files referenced here
            ...
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            ...
        self.object_map[record.level-1].resourceType = ResourceType.DigitalArtifact
           
    def handle_form(self, record: Gedcom5xRecord):
        if record.parent is not None and record.parent.tag == 'FILE' and isinstance(self.object_map[record.level-2], SourceDescription):
            if record.value and record.value.strip() != '':
                mime_type, _ = mimetypes.guess_type('placehold.' + record.value)
                if mime_type:
                    self.object_map[record.level-2].mediaType = mime_type
                else:
                    log.error(f"Could not determing mime type from {record.value}")
        elif isinstance(self.object_map[record.level-1], PlaceDescription):
            self.object_map[record.level-1].names.append(TextValue(value=record.value))
        elif record.parent is not None and record.parent.tag == 'TRAN':
            pass #TODO
        else:
            log.error(f"raise TagConversionError(record=record,levelstack=self.object_map")

    def handle_givn(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Name):
            given_name = NamePart(value=record.value, type=NamePartType.Given)
            self.object_map[record.level-1]._add_name_part(given_name)
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_indi(self, record: Gedcom5xRecord):
        person = self.gedcomx.persons.byId(record.xref)
        if person is None:
            log.warning('Had to create person with id {recrod.xref}')
            if isinstance(record.xref,str):
                person = Person(id=record.xref)
                self.gedcomx.add_person(person)
            else:
                raise ValueError('INDI Record had no XREF')    
        self.object_map[record.level] = person

    def handle_immi(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Immigration)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_marr(self, record: Gedcom5xRecord):
        """
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Marriage)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        """
        if (add_fact := getattr(self.object_map[record.level-1],'add_fact',None)) is not None:
            gxobject = Fact(type=FactType.Marriage)
            add_fact(gxobject)
            self.object_map[record.level] = gxobject       
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle__link(self,record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceReference):
            gxobject = Identifier([URI.from_url(record.value)],IdentifierType.External)
            self.object_map[record.level-1].description.add_identifier(gxobject)
            self.object_map[record.level] = gxobject

    def handle__milt(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.MilitaryService)
            self.object_map[record.level-1].add_fact(gxobject)
            self.object_map[record.level] = gxobject

    def handle_name(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Name.simple(record.value if record.value else 'WARNING: NAME had no value')
            #gxobject = Name(nameForms=[NameForm(fullText=record.value)], type=NameType.BirthName)
            self.object_map[record.level-1].add_name(gxobject)

            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Agent):
            
            gxobject = TextValue(value=record.value)
            self.object_map[record.level-1].add_name(gxobject)
            
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_note(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            gxobject = Note(text=self.clean_str(record.value))
            self.object_map[record.level-1].add_note(gxobject)

            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], SourceReference):
            if self.object_map[record.level-1].description is not None:
                gxobject = Note(text=self.clean_str(record.value))
                self.object_map[record.level-1].description.add_note(gxobject)
                self.object_map[record.level] = gxobject
            else:
                log.error('SourceReference does not have description')

            
            
        elif isinstance(self.object_map[record.level-1], Conclusion):
            gxobject = Note(text=record.value)
            self.object_map[record.level-1].add_note(gxobject)

            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Agent):
            gxobject = Note(text=record.value)
            self.object_map[record.level-1].add_note(gxobject)

            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Attribution):
            if self.object_map[record.level-1].changeMessage is None:
                self.object_map[record.level-1].changeMessage = record.value
            else:
                self.object_map[record.level-1].changeMessage = self.object_map[record.level-1].changeMessage + '' + record.value
        elif isinstance(self.object_map[record.level-1], Note):
            gxobject = Note(text=self.clean_str(record.value))
            self.object_map[record.level-2].add_note(gxobject)

            
            self.object_map[record.level] = gxobject

        else:
            raise ValueError(f"Could not handle 'NOTE' tag in record {record.describe()}, last stack object {type(self.object_map[record.level-1])}")
            assert False

    def handle_nsfx(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Name):
            surname = NamePart(value=record.value, type=NamePartType.Suffix)
            self.object_map[record.level-1]._add_name_part(surname)
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_occu(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Occupation)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_obje(self, record: Gedcom5xRecord):
        self.handle_sour(record)

    def handle_page(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceReference):
            #self.object_map[record.level-1].descriptionId = record.value
            gx_object = KnownSourceReference(name=KnownSourceReference.Page,value=record.value)
            self.object_map[record.level-1].add_qualifier(gx_object)
            self.object_map[record.level] = self.object_map[record.level-1]
        else:
            raise ValueError(f"Could not handle 'PAGE' tag in record {record.describe()}, object stack {self.object_map}")

    def handle_plac(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            gxobject = Address(value=record.value)
            self.object_map[record.level-1].add_address(gxobject)

            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Event):
            if self.gedcomx.places.byName(record.value):
                self.object_map[record.level-1].place = PlaceReference(original=record.value, description=self.gedcomx.places.byName(record.value)[0])
            else:
                place_des = PlaceDescription(names=[TextValue(value=record.value)])
                self.gedcomx.add_place_description(place_des)
                self.object_map[record.level-1].place = PlaceReference(original=record.value, description=place_des)
                if len(record.subRecords()) > 0:
                    self.object_map[record.level]= place_des

        elif isinstance(self.object_map[record.level-1], Fact):
            if self.gedcomx.places.byName(record.value):
                self.object_map[record.level-1].place = PlaceReference(original=record.value, description=self.gedcomx.places.byName(record.value)[0])
            else:
                place_des = PlaceDescription(names=[TextValue(value=record.value)])
                self.gedcomx.add_place_description(place_des)
                self.object_map[record.level-1].place = PlaceReference(original=record.value, description=place_des)
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            gxobject = Note(text='Place: ' + record.value if record.value else 'WARNING: NOTE had no value')
            self.object_map[record.level-1].add_note(gxobject)
            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_post(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            self.object_map[record.level-1].postalCode = self.clean_str(record.value)
        else:
            raise ValueError(f"I do not know how to handle an 'POST' tag for a {type(self.object_map[record.level-1])}")   
    
    def handle_publ(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            if record.value and self.gedcomx.agents.byName(record.value):
                gxobject = self.gedcomx.agents.byName(record.value)[0]
            else:
                gxobject = Agent(names=[TextValue(record.value)])
                self.gedcomx.add_agent(gxobject)
            self.object_map[record.level-1].publisher = gxobject

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_prob(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Probate)
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_uid(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            gxobject = Identifier(value=[URI('UID:' + record.value)] if record.value else [URI('WARNING: NOTE had no value')],type=IdentifierType.Primary) # type: ignore
            self.object_map[record.level-1].add_identifier(gxobject) #NOTE GC7
            
            self.object_map[record.level] = gxobject

    def handle_refn(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person) or isinstance(self.object_map[record.level-1], SourceDescription):
            gxobject = Identifier(value=[URI.from_url('Reference Number:' + record.value)] if record.value else [])
            self.object_map[record.level-1].add_identifier(gxobject)
            
            self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], Agent):
            gxobject = Identifier(value=[URI('Reference Number:' + record.value)] if record.value else [])
            self.object_map[record.level-1].add_identifier(gxobject) #NOTE GC7
            
            self.object_map[record.level] = gxobject
        else:
            raise ValueError(f"Could not handle 'REFN' tag in record {record.describe()}, last stack object {type(self.object_map[record.level-1])}")

    def handle_repo(self, record: Gedcom5xRecord):

        if record.level == 0:
            if (gxobject := self.gedcomx.agents.byId(id=record.xref)) is None:
                gxobject = Agent(id=record.xref)
                self.gedcomx.add_agent(gxobject)
            
            self.object_map[record.level] = gxobject
            
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            if self.gedcomx.agents.byId(record.xref) is not None:
                
                # TODO WHere and what to add this to?
                gxobject = self.gedcomx.agents.byId(record.xref)
                self.object_map[record.level-1].repository = gxobject
                self.object_map[record.level] = gxobject

            else:
                raise ValueError()
                gxobject = Agent(names=[TextValue(record.value)])
        else:
            raise ValueError(f"I do not know how to handle 'REPO' tag that is not a top-level, or sub-tag of {type(self.object_map[record.level-1])}")
            

        
        self.object_map[record.level] = gxobject

    def handle_resi(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Person):
            gxobject = Fact(type=FactType.Residence)
            if record.value and record.value.strip() != '':
                gxobject.add_note(Note(text=record.value))
            self.object_map[record.level-1].add_fact(gxobject)

            
            self.object_map[record.level] = gxobject
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_rin(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            self.object_map[record.level-1].id = record.value
            self.object_map[record.level-1].add_note(Note(text=f"Source had RIN: of {record.value}"))

        else:
            raise ValueError(f"Could not handle 'RIN' tag in record {record.describe()}, last stack object {type(self.object_map[record.level-1])}")
        
    def handle_sex(self, record: Gedcom5xRecord):
        
        if isinstance(self.object_map[record.level-1], Person):
            if record.value == 'M':
                gxobject = Gender(type=GenderType.Male)
            elif record.value == 'F':
                gxobject = Gender(type=GenderType.Female)
            else:
                gxobject = Gender(type=GenderType.Unknown)
            self.object_map[record.level-1].gender = gxobject
            
            
            self.object_map[record.level] = gxobject
        else:
            assert False

    def handle_sour(self, record: Gedcom5xRecord):
        if record.level == 0 and (record.tag in ['SOUR','OBJE','_WLNK']):
            
            if (gxobject := self.gedcomx.sourceDescriptions.byId(record.xref)) is None:
                log.debug(f"SourceDescription with id: {record.xref} was not found. Creating a new SourceDescription")
                log.debug(f"Creating SourceDescription from {record.tag} {record.describe()}")
                gxobject = SourceDescription(id=record.xref)
                self.object_map[record.level-1].add_source_description(gxobject)
            else:
                log.debug(f"Found SourceDescription with id:{record.xref}")
            
        elif (add_method := getattr(self.object_map[record.level-1],"add_source_reference",None)) is not None:
            if (source_description := self.gedcomx.sourceDescriptions.byId(record.xref)) is not None:
                gxobject = SourceReference(descriptionId=record.xref, description=source_description)
                add_method(gxobject)
            else:
                log.error(f"Could not find source with id: {record.xref}, Creating Place Holder Description")
                gxobject = SourceDescription(id=record.xref)
                gxobject._place_holder = True
                gxobject = SourceReference(descriptionId=record.xref, description=gxobject)

        elif record.tag == 'OBJE' and isinstance(self.object_map[record.level-1],SourceReference):
            if (source_description := self.gedcomx.sourceDescriptions.byId(record.xref)) is not None:
                gxobject = SourceReference(descriptionId=record.xref, description=source_description)
                self.object_map[record.level-1].description.add_source_reference(gxobject)
            else:
                assert False

        else:
            print(record.describe())
            print(self.object_map)
            assert False
            
        self.object_map[record.level] = gxobject
          
    def handle_stae(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Address):
            self.object_map[record.level-1].stateOrProvince = self.clean_str(record.value)
        else:
            raise ValueError(f"I do not know how to handle an 'STAE' tag for a {type(self.object_map[record.level-1])}")
        
    def handle_surn(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Name):
            surname = NamePart(value=record.value, type=NamePartType.Surname)
            self.object_map[record.level-1]._add_name_part(surname)
        else:
            raise TagConversionError(record=record,levelstack=self.object_map)

    def handle_text(self, record: Gedcom5xRecord):
        if record.parent is not None and record.parent.tag == 'DATA':
            if isinstance(self.object_map[record.level-2], SourceReference):
                gxobject = TextValue(value=record.value)
                self.object_map[record.level-2].description.add_description(gxobject)
                
                self.object_map[record.level] = gxobject
        elif isinstance(self.object_map[record.level-1], SourceDescription):
            gxobject = Document(text=record.value)
            self.object_map[record.level-1].analysis = gxobject
        else:
            assert False

    def handle_titl(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], SourceDescription):
            
            gxobject = TextValue(value=self.clean_str(record.value))
            self.object_map[record.level-1].add_title(gxobject)
            self.object_map[record.level] = gxobject
        
        elif record.parent is not None and record.parent.tag == 'FILE' and isinstance(self.object_map[record.level-2], SourceDescription):
            gxobject = TextValue(value=record.value)
            self.object_map[record.level-2].add_title(gxobject)

            
            self.object_map[record.level] = gxobject
        elif self.object_map[record.level] and isinstance(self.object_map[record.level], Name):
            gxobject = NamePart(value=record.value, qualifiers=[NamePartQualifier.Title])

            self.object_map[record.level]._add_name_part(gxobject)
        else:
            log.error(f"raise TagConversionError(record=record,levelstack=self.object_map)")

    def handle_tran(self, record: Gedcom5xRecord):
        pass

    def handle_type(self, record: Gedcom5xRecord):
        # peek to see if event or fact
        if isinstance(self.object_map[record.level-1], Event):
            if EventType.guess(record.value):
                self.object_map[record.level-1].type = EventType.guess(record.value)                
            else:
                log.warning(f"Could not determine type of event with value '{record.value}'")  
            # add as a note anyway, guess works of text in the string    
            self.object_map[record.level-1].add_note(Note(text=self.clean_str(record.value)))
        elif isinstance(self.object_map[record.level-1], Fact):
            if not self.object_map[record.level-1].type:
                self.object_map[0].type = FactType.guess(record.value)
        elif isinstance(self.object_map[record.level-1], Identifier):
            
            self.object_map[record.level-1].values.append(self.clean_str(record.value))
            self.object_map[record.level-1].type = IdentifierType.Other # type: ignore

        elif record.parent is not None and record.parent.tag == 'FORM':
            if not self.object_map[0].mediaType:
                self.object_map[0].mediaType = record.value
        elif isinstance(self.object_map[record.level-1], Name):
            self.object_map[record.level-1].type = GedcomConverter.type_name_type.get(record.value,NameType.Other)

        else:
            raise TagConversionError(record,self.object_map)

    def handle__url(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-2], SourceDescription):
            self.object_map[record.level-2].about = URI.from_url(record.value) if record.value else None
        else:
            raise ValueError(f"Could not handle '_URL' tag in record {record.describe()}, last stack object {self.object_map[record.level-1]}")
            
    def handle_www(self, record: Gedcom5xRecord):
        if isinstance(self.object_map[record.level-1], Agent):
            self.object_map[record.level-1].homepage = self.clean_str(record.value)
        elif isinstance(self.object_map[record.level-2], SourceReference):
            self.object_map[record.level-2].description.add_identifier(Identifier(value=[URI.from_url(record.value)] if record.value else []))
        else:
            raise ValueError(f"Could not handle 'WWW' tag in record {record.describe()}, last stack object {self.object_map[record.level-1]}")

    def parse_gedcom5x_fam_record(self, record: Gedcom5xRecord):
        log.info(f"Parsing family recrods")
        with open('./logs/gedcomx.convert.families.json', 'a') as f:
            for fam in record._flatten_subrecords(record):
                f.write(fam.describe() + "\n")

    def print_counts_table(self, counts: Mapping[Any, int]) -> None:
        """
        Pretty-print {key: int} as columns, largest count first.
        Column count adapts to terminal width and number of items.
        """
        items = [(str(k), int(v)) for k, v in counts.items()]
        if not items:
            print("(empty)")
            return

        # Sort: by value desc, then key asc for stable ordering
        items.sort(key=lambda kv: (-kv[1], kv[0]))

        # Cell formatting widths
        key_w = max(len(k) for k, _ in items)
        num_w = max(len(str(v)) for _, v in items)
        cell_fmt = f"{{k:<{key_w}}}  {{v:>{num_w}}}"  # e.g., 'Surname        123'
        cell_width = key_w + 2 + num_w + 2            # +2 padding between columns

        # Decide number of columns: fit to terminal, but also scale with item count
        term_cols = shutil.get_terminal_size(fallback=(100, 24)).columns
        fit_cols = max(1, term_cols // cell_width)
        sqrt_cols = max(1, int(math.sqrt(len(items))))  # more cols when many items
        cols = max(1, min(len(items), max(fit_cols, sqrt_cols)))

        rows = math.ceil(len(items) / cols)

        # Print row-wise, reading items column-major so columns stay balanced
        for r in range(rows):
            line = []
            for c in range(cols):
                i = c * rows + r
                if i < len(items):
                    k, v = items[i]
                    cell = cell_fmt.format(k=k, v=v)
                    line.append(cell.ljust(cell_width))
            print("".join(line).rstrip())

    def Gedcom5x_GedcomX(self, gedcom5x: Gedcom5x):
        print(f'Parsing GEDCOM Version {gedcom5x.version}')
        individual_ids = set()
        source_ids = set()
        repository_ids = set()
        family_ids = set()

        with hub.use(convert_log):
            if gedcom5x:
                for object in gedcom5x.objects:
                    log.debug(f"Priming Source id's with id: {object.xref} from objects")
                    source_ids.add(object.xref)
                    gx_obj = SourceDescription(id=object.xref)
                    self.gedcomx.add_source_description(gx_obj)

                for source in gedcom5x.sources:
                    source_ids.add(source.xref)
                    gx_obj = SourceDescription(id=source.xref)
                    self.gedcomx.add_source_description(gx_obj)
                
                for repo in gedcom5x.repositories:
                    repository_ids.add(repo.xref)
                    gx_obj = Agent(id=repo.xref)
                    self.gedcomx.add_agent(gx_obj)
                
                for individual in gedcom5x.individuals:
                    individual_ids.add(individual.xref)
                    gx_obj = Person(id=individual.xref)
                    self.gedcomx.add_person(gx_obj)
                
                for family in gedcom5x.families:
                    family_ids.add(family.xref)
                    self.handle_fam(family)
                
                # Now Parse Zero Level Recrods
                for source in gedcom5x.sources:
                    self.parse_gedcom5x_record(source)
                for object in gedcom5x.objects:
                    self.parse_gedcom5x_record(object)
                for individual in gedcom5x.individuals:
                    self.parse_gedcom5x_record(individual)
                for repo in gedcom5x.repositories:
                    self.parse_gedcom5x_record(repo)
                for family in gedcom5x.families:
                    self.parse_gedcom5x_record(family)
        
        self.print_counts_table(self.missing_handler_count)
        
        return self.gedcomx



from typing import List, Optional
from urllib.parse import urljoin

"""
======================================================================
 Project: Gedcom-X
 File:    Person.py
 Author:  David J. Cartwright
 Purpose: Python Object representation of GedcomX Person Type

 Created: 2025-08-25
 Updated:
   - 2025-08-31: _as_dict_ to only create entries in dict for fields that hold data
   - 2025-09-03: _from_json_ refactor
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .attribution import Attribution
from .conclusion import ConfidenceLevel
from .date import Date
from .evidence_reference import EvidenceReference
from .Extensions.rs10.rsLink import _rsLinks
from .extensible import Extensible
from .fact import Fact, FactType
from .gender import Gender, GenderType
from .identifier import IdentifierList
from .logging_hub import hub, logging
from .name import Name, QuickName
from .note import Note
from .resource import Resource
from .source_reference import SourceReference
from .subject import Subject
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
deserial_log = "degedcomx.serialization"
#=====================================================================



class Person(Extensible,Subject):
    """A person in the system.

    Args:
        id (str):      Unique identifier for this person.
        name (str):    Full name.
        birth (date):  Birth date (YYYY-MM-DD).
        friends (List[Person], optional): List of friends. Defaults to None.

    Raises:
        ValueError: If `id` is not a valid UUID.
    """
    identifier = 'http://gedcomx.org/v1/Person'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, id: str | None = None,
             lang: str = 'en',
             sources: Optional[List[SourceReference]] = None,
             analysis: Optional[Resource] = None,
             notes: Optional[List[Note]] = None,
             confidence: Optional[ConfidenceLevel] = None,
             attribution: Optional[Attribution] = None,
             extracted: Optional[bool] = None,
             evidence: Optional[List[EvidenceReference]] = None,
             media: Optional[List[SourceReference]] = None,
             identifiers: Optional[IdentifierList] = None,
             private: Optional[bool] = None,
             gender: Optional[Gender] = Gender(type=GenderType.Unknown),
             names: Optional[List[Name]] = None,
             facts: Optional[List[Fact]] = None,
             living: Optional[bool] = None,
             links: Optional[_rsLinks] = None,
             uri: Optional[Resource] = None) -> None:
        # Call superclass initializer if needed
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution, extracted, evidence, media, identifiers,links=links,uri=uri)
        
        # Initialize mutable attributes to empty lists if None
        self.sources = sources if sources is not None else []
        self.notes = notes if notes is not None else []
        self.evidence = evidence if evidence is not None else []
        self.media = media if media is not None else []
        self.identifiers = identifiers if identifiers is not None else []
        self.names = names if names is not None else []
        self.facts = facts if facts is not None else []

        self.private = private
        self.gender = gender

        self.living = living       #TODO This is from familysearch API

        self._relationships = []
          
    def add_fact(self, fact_to_add: Fact) -> bool:
        if fact_to_add and isinstance(fact_to_add,Fact):
            for current_fact in self.facts:
                if fact_to_add == current_fact:
                    return False
            self.facts.append(fact_to_add)
            return True
        return False

    def add_name(self, name_to_add: Name) -> bool:
        if len(self.names) > 5: 
            for name in self.names:
                print(name)
            raise
        if name_to_add and isinstance(name_to_add, Name):
            for current_name in self.names:
                if name_to_add == current_name:
                    return False
            self.names.append(name_to_add)
            return True
        return False
    
    def _add_relationship(self, relationship_to_add: object):
        from .relationship import Relationship
        if isinstance(relationship_to_add,Relationship):
            self._relationships.append(relationship_to_add)
        else:
            raise ValueError()
    
    def display(self):
        display = {
        "ascendancyNumber": "1",
        "deathDate": "from 2001 to 2005",
        "descendancyNumber": "1",
        "gender": self.gender.type if self.gender else 'Unknown',
        "lifespan": "-2005",
        "name": self.names[0].nameForms[0].fullText
            }
        
        return display

    @property
    def _as_dict_(self):
        from .serialization import Serialization
        return Serialization.serialize(self)
                
           
    @classmethod
    def _from_json_(cls, data: dict, context = None) -> "Person":
        def _to_bool(v):
            if isinstance(v, str):
                return v.strip().lower() in ("1", "true", "yes", "y", "t")
            return bool(v)
        with hub.use(deserial_log):
            log.debug(f"Deserializing a Person")
            person_data: dict = Subject._dict_from_json_(data,context)
            if (private := data.get("private")) is not None:
                person_data["private"] = _to_bool(private)
            if (living := data.get("living")) is not None:
                person_data["living"] = _to_bool(living)
            if (gender := data.get("gender")) is not None:
                person_data["gender"] = Gender._from_json_(gender, context)
            if (names := data.get("names")) is not None:
                person_data["names"] = [Name._from_json_(n, context) for n in names]
            if (facts := data.get("facts")) is not None:
                person_data["facts"] = [Fact._from_json_(f, context) for f in facts]
            
            diff = data.keys() - person_data.keys()
            log.debug(f"Desserialization found keys {diff} that are not in deserialization list")

        return cls(**person_data)

    @classmethod
    def from_familysearch(cls, pid: str, token: str, *, base_url: Optional[str] = None):
        from .serialization import Serialization
        """
        Fetch a single person by PID from FamilySearch and return a Person.
        - pid: e.g. "KPHP-4B4"
        - token: OAuth2 access token (Bearer)
        - base_url: override API base (defaults to settings.FS_API_BASE or prod)
        """
        import requests
        default_base = "https://apibeta.familysearch.org/platform/"

        base = (base_url or default_base).rstrip("/") + "/"
        url = urljoin(base, f"tree/persons/{pid}")

        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {token}",
        }

        resp = requests.get(url, headers=headers, timeout=(5, 30))
        resp.raise_for_status()

        payload = resp.json()
        persons = payload.get("persons") or []

        # Prefer exact match on PID, else first item if present
        person_json = next((p for p in persons if (p.get("id") == pid)), None) or (persons[0] if persons else None)
        if not person_json:
            raise ValueError(f"FamilySearch returned no person for PID {pid}")

        # Keep your existing deserialization helper
        return Person._from_json_(person_json)    

class QuickPerson:
    """A GedcomX Person Data Type created with basic information.

    Underlying GedcomX Types are created for you.
        
    Args:
        name (str):    Full name.
        birth (date,Optional):  Birth date (YYYY-MM-DD).
        death (date, Optional)

        

    Raises:
        ValueError: If `id` is not a valid UUID.
    """
    def __new__(cls, name: str, dob: Optional[str] = None, dod: Optional[str] = None):
        # Build facts from args
        facts = []
        if dob:
            facts.append(Fact(type=FactType.Birth, date=Date(original=dob)))
        if dod:
            facts.append(Fact(type=FactType.Death, date=Date(original=dod)))

        # Return the different class instance
        return Person(facts=facts, names=[QuickName(name=name)] if name else None)


from __future__ import annotations
from dataclasses import dataclass, field
from typing import Mapping, Sequence, Tuple, Union, Iterable
from urllib.parse import urlsplit, urlunsplit, urlencode, parse_qsl, SplitResult
from urllib.parse import urlunparse

"""
======================================================================
 Project: Gedcom-X
 File:    uri.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactor 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

_DEFAULT_SCHEME = "gedcomx"

# ---------- typing helpers for urlencode (Option A) ----------
QuerySeq = Sequence[Tuple[str, str]]
QueryMap = Mapping[str, str]
QueryLike = Union[str, QueryMap, QuerySeq]

def _encode_query(q: QueryLike) -> str:
    """Return a properly encoded query string."""
    if isinstance(q, str):
        return q
    if isinstance(q, Mapping):
        return urlencode(q, doseq=True)          # mapping is fine
    return urlencode(list(q), doseq=True)        # coerce iterable to a sequence


class URI():
    def __init__(self,
                 
                 target=None,
                 scheme: str | None = None,
                 authority: str | None = None,
                 path: str | None = None,
                 params: str | None = None,
                 query: str | None = None,
                 fragment: str | None = None,
                 value: str | None = None
                 ) -> None:
        
        self.target = target

        self.scheme = scheme 
        self.authority = authority
        self.path = path
        self.params = params
        self.query = query
        self.fragment = fragment     
        
        self._value = value

        if self._value:
            s = urlsplit(self._value)
            self.scheme = s.scheme or _DEFAULT_SCHEME
            self.authority=s.netloc
            self.path=s.path
            self.query=s.query
            self.fragment=s.fragment

        if self.target is not None:
            #log.debug(f"Creating URI from Target {target}, most likely for serialization")
            if hasattr(self.target,'id'):
                log.debug(f"'{target}.id' found {target.id}, using as fragment")
                self.fragment = self.target.id
            if hasattr(self.target,'uri'):
                #log.debug(f"'{target}.uri' found, copying")
                self._value = target.uri._value
                self.scheme = target.uri.scheme
                self.authority = target.uri.authority 
                self.path = target.uri.path 
                self.query = target.uri.query
                self.fragment = target.uri.fragment
            elif isinstance(target,URI):
                #log.debug(f"'{target} is a URI, copying")
                self._value = target._value
                self.scheme = target.scheme
                self.authority = target.authority 
                self.path = target.path 
                self.query = target.query
                self.fragment = target.fragment
            
            
            
            elif isinstance(self.target,str):
                #log.warning(f"Creating a URI from target type {type(target)} with data: {target}.")
                s = urlsplit(self.target)
                self.scheme = s.scheme or _DEFAULT_SCHEME
                self.authority=s.netloc
                self.path=s.path
                self.query=s.query
                self.fragment=s.fragment
            else:
                #log.warning(f"Unable to create URI from target type {type(target)} with data: {target}.")
                self._value = target
        #log.info(f"self.scheme = {self.scheme} self.authority={self.authority} self.path={self.path} self.query={self.query}  self.fragment={self.fragment}")

        parts = [
        self.scheme or "",
        self.authority or "",
        self.path or "",
        self.params or "",
        self.query or "",
        self.fragment or "",
        ]
        if not any(parts):
            raise ValueError()   

    @property
    def value(self) -> str:
        parts = [
        self.scheme or "",
        self.authority or "",
        self.path or "",
        self.params or "",
        self.query or "",
        self.fragment or "",
        ]
        if not any(parts):
            return None
        return str(urlunparse(parts))

    def split(self) -> SplitResult:
        return SplitResult(self.scheme, self.authority, self.path, self.query, self.fragment)

    def __str__(self) -> str:
        return urlunsplit(self.split())
    
    def __repr__(self) -> str:
        return (f"scheme = {self.scheme}, authority={self.authority}, path={self.path}, query={self.query}, fragment={self.fragment}")
    
    @property
    def _as_dict_(self):
        return self.value or self._value    
    
    @property
    def uri(self):
        return self
    
    @classmethod
    def from_url(cls,url):
        return cls(target=url)
    
    @classmethod
    def _from_json_(cls,data,context=None):
        return cls(value=data)
    

    
@dataclass(slots=True)
class _URI:
    scheme: str = field(default=_DEFAULT_SCHEME)
    authority: str = field(default="")
    path: str = field(default="")
    query: str = field(default="")
    fragment: str = field(default="")

    # ---------- constructors ----------
    @classmethod
    def from_url(cls, url: str, *, default_scheme: str = _DEFAULT_SCHEME) -> URI:
        s = urlsplit(url)
        scheme = s.scheme or default_scheme
        return cls(scheme=scheme, authority=s.netloc, path=s.path, query=s.query, fragment=s.fragment)

    @classmethod
    def parse(cls, value: str) -> URI:
        return cls.from_url(value)

    @classmethod
    def from_parts(
        cls,
        *,
        scheme: str | None = None,
        authority: str = "",
        path: str = "",
        query: QueryLike = "",
        fragment: str = "",
    ) -> URI:
        q = _encode_query(query)
        return cls(scheme=scheme or _DEFAULT_SCHEME, authority=authority, path=path, query=q, fragment=fragment)

    # ---------- views ----------
    @property
    def uri(self) -> str:
        return str(self)
    
    @property
    def value(self) -> str:
        return str(self)

    def split(self) -> SplitResult:
        return SplitResult(self.scheme, self.authority, self.path, self.query, self.fragment)

    def __str__(self) -> str:
        return urlunsplit(self.split())

    @property
    def _as_dict_(self) -> dict[str, object]:
        return {
            "scheme": self.scheme,
            "authority": self.authority,
            "path": self.path,
            "query": self.query,
            "fragment": self.fragment,
            "value": str(self),
        }

    # Accepts {'resource': '...'} or a plain string
    @classmethod
    def _from_json_(cls, data: str | Mapping[str, object],context=None) -> URI:
        return cls.from_parts(fragment="NOT IMPLIMENTED")
        if isinstance(data, str):
            return cls.from_parts(fragment="NOT IMPLIMENTED")
        if isinstance(data, Mapping):
            raw = data.get("resource") or data.get("value") or ""
            if isinstance(raw, str) and raw:
                return cls.from_url(raw)
        raise ValueError(f"Cannot build URI from: {data!r}")

    # ---------- functional updaters ----------
    def with_scheme(self, scheme: str) -> URI: return self.replace(scheme=scheme)
    def with_authority(self, authority: str) -> URI: return self.replace(authority=authority)
    def with_path(self, path: str, *, join: bool = False) -> URI:
        new_path = (self.path.rstrip("/") + "/" + path.lstrip("/")) if join else path
        return self.replace(path=new_path)
    def with_fragment(self, fragment: str | None) -> URI:
        return self.replace(fragment=(fragment or ""))
    def without_fragment(self) -> URI: return self.replace(fragment="")
    def with_query(self, query: QueryLike) -> URI:
        return self.replace(query=_encode_query(query))
    def add_query_params(self, params: Mapping[str, Union[str, Iterable[str]]]) -> URI:
        existing = parse_qsl(self.query, keep_blank_values=True)
        for k, v in params.items():
            if isinstance(v, str):
                existing.append((k, v))
            else:
                for vv in v:
                    existing.append((k, vv))
        return self.replace(query=urlencode(existing, doseq=True))

    # ---------- helpers ----------
    def replace(self, **kwargs) -> URI:
        cls = type(self)
        return cls(
            scheme=kwargs.get("scheme", self.scheme or _DEFAULT_SCHEME),
            authority=kwargs.get("authority", self.authority),
            path=kwargs.get("path", self.path),
            query=kwargs.get("query", self.query),
            fragment=kwargs.get("fragment", self.fragment),
        )



# setup.py
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='gedcom-x',
    version='0.5.10',
    packages=find_packages(),
    description="A Python toolkit for working with GEDCOM-X",
    long_description=long_description,
    long_description_content_type="text/markdown",
    #install_requires=[
        # List your project dependencies here, e.g.,
    #    None,
    #],
    entry_points={
        'console_scripts': [
            # Define command-line scripts if needed
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)


# extensible.py
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Callable

from .schemas import SCHEMA
# If you're using the earlier schema registry:
#from .serialization_schema import SCHEMA



class Extensible:
    # class-level registry of declared extras
    _declared_extras: Dict[str, Any] = {}

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        # each subclass gets its own dict (copy, not shared)
        cls._declared_extras = dict(getattr(cls, "_declared_extras", {}))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # cooperative
        self.extras: Dict[str, Any] = {}
        # seed declared defaults
        for k, default in type(self)._declared_extras.items():
            self.extras[k] = _copy_default(default)

    @classmethod
    def define_ext(
        cls,
        name: str,
        *,
        typ: type | None = None,
        default: Any = None,
        overwrite: bool = False,
    ) -> None:
        """
        Declare an extra field on the CLASS.

        Args:
            name: field name
            typ: Python type (used to update schema registry)
            default: default value for new instances
            overwrite: if True, replaces existing definition
        """
        if name in getattr(cls, "__dataclass_fields__", {}):
            raise AttributeError(f"{name!r} already exists on {cls.__name__}")

        already = hasattr(cls, name)
        if already and not overwrite:
            return

        # Attach descriptor
        setattr(cls, name, _ExtraField(name, default))
        cls._declared_extras[name] = default

        # Register with schema
        if typ is None and default is not None:
            typ = type(default)
        SCHEMA.register_extra(cls, name, typ or type(None))

    @classmethod
    def declared_extras(cls) -> Dict[str, Any]:
        return dict(getattr(cls, "_declared_extras", {}))


class _ExtraField:
    def __init__(self, name: str, default: Any):
        self.name = name
        self.default = default
    def __get__(self, obj, owner):
        if obj is None:
            return self
        return obj.extras.get(self.name, self.default)
    def __set__(self, obj, value):
        obj.extras[self.name] = value


def _copy_default(v: Any) -> Any:
    if isinstance(v, (list, dict, set)):
        return v.copy()
    return v
    # avoid shared mutable defaults
    if isinstance(v, (list, dict, set)):
        return v.copy()
    return v


from typing import Any, Dict, List, Optional

"""
======================================================================
 Project: Gedcom-X
 File:    place_description.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-01: filename PEP8 standard
   - 2025-09-03: _from_json_ refactored
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .attribution import Attribution
from .conclusion import ConfidenceLevel
from .date import Date
from .evidence_reference import EvidenceReference
from .Extensions.rs10.rsLink import _rsLinks #new
from .identifier import IdentifierList
from .note import Note
from .resource import Resource
from .source_reference import SourceReference
from .subject import Subject
from .textvalue import TextValue
from .uri import URI
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class PlaceDescription(Subject):
    """PlaceDescription describes the details of a place in terms of 
    its name and possibly its type, time period, and/or a geospatial description
    functioning as a description of a place as a snapshot in time.

    Encapsulates textual names, geospatial coordinates, jurisdictional context,
    temporal coverage, and related resources (media, sources, evidence, etc.).
    

    Attributes:
        names (Optional[List[TextValue]]): Human-readable names or labels for
            the place (e.g., “Boston, Suffolk, Massachusetts, United States”).
        type (Optional[str]): A place type identifier (e.g., a URI). **TODO:**
            replace with an enumeration when finalized.
        place (Optional[URI]): Canonical identifier (URI) for the place.
        jurisdiction (Optional[Resource|PlaceDescription]): The governing or
            containing jurisdiction of this place (e.g., county for a town).
        latitude (Optional[float]): Latitude in decimal degrees (WGS84).
        longitude (Optional[float]): Longitude in decimal degrees (WGS84).
        temporalDescription (Optional[Date]): Temporal coverage/validity window
            for this description (e.g., when a jurisdictional boundary applied).
        spatialDescription (Optional[Resource]): A resource describing spatial
            geometry or a link to an external gazetteer/shape definition.
    """
    identifier = "http://gedcomx.org/v1/PlaceDescription"
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, id: Optional[str] =None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[Resource] = None,
                 notes: Optional[List[Note]] =None,
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None,
                 extracted: Optional[bool] = None,
                 evidence: Optional[List[EvidenceReference]] = None,
                 media: Optional[List[SourceReference]] = None,
                 identifiers: Optional[IdentifierList] = None,
                 names: Optional[List[TextValue]] = None,
                 type: Optional[str] = None,    #TODO This needs to be an enumerated value, work out details
                 place: Optional[URI] = None,
                 jurisdiction: Optional["Resource | PlaceDescription"] = None, 
                 latitude: Optional[float] = None,
                 longitude: Optional[float] = None,
                 temporalDescription: Optional[Date] = None,
                 spatialDescription: Optional[Resource] = None,
                 links: Optional[_rsLinks] = None
                 ) -> None:
        
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution, extracted, evidence, media, identifiers,links=links)
        self.names = names
        self.type = type
        self.place = place
        self.jurisdiction = jurisdiction
        self.latitude = latitude
        self.longitude = longitude
        self.temporalDescription = temporalDescription
        self.spatialDescription = spatialDescription

    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = super()._as_dict_ or {}
        
        if self.names:
            type_as_dict["names"] = [n._as_dict_ for n in self.names if n]
        if self.type:
            type_as_dict["type"] = self.type    #TODO
        if self.place:
            type_as_dict["place"] = self.place._as_dict_
        if self.jurisdiction:
            type_as_dict["jurisdiction"] = self.jurisdiction._as_dict_ 
        if self.latitude is not None: # include 0.0; exclude only None
            type_as_dict["latitude"] = float(self.latitude)
        if self.longitude is not None: # include 0.0; exclude only None
            type_as_dict["longitude"] = float(self.longitude)
        if self.temporalDescription:
            type_as_dict["temporalDescription"] = self.temporalDescription._as_dict_
        if self.spatialDescription:
            type_as_dict["spatialDescription"] = self.spatialDescription._as_dict_

        return type_as_dict if type_as_dict != {} else None
        return Serialization.serialize_dict(type_as_dict) 

    @classmethod
    def _from_json_(cls, data: Any, context: Any = None) -> "PlaceDescription":
        """
        Create a PlaceDescription instance from a JSON-dict (already parsed).
        """        
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)}")

        person_data: Dict[str, Any] = Subject._dict_from_json_(data,context)

        # names (allow both list and a single 'name' alias)
        if (names := data.get("names")) is not None:
            person_data["names"] = [TextValue._from_json_(n, context) for n in names]
        
        # type (string for now; promote to enum later)
        if (typ := data.get("type")) is not None:
            person_data["type"] = typ

        # place: URI (accept string or dict)
        if (pl := data.get("place")) is not None:
            person_data["place"] = URI(pl)

        # jurisdiction: Resource | PlaceDescription
        if (jur := data.get("jurisdiction")) is not None:
            person_data["jurisdiction"] = Resource._from_json_(jur, context)
            
        # coordinates
        if (lat := data.get("latitude")) is not None:
            person_data["latitude"] = float(lat)
        if (lon := data.get("longitude")) is not None:
            person_data["longitude"] = float(lon)

        # temporal / spatial descriptions
        if (td := data.get("temporalDescription")) is not None:
            person_data["temporalDescription"] = Date._from_json_(td, context)

        if (sd := data.get("spatialDescription")) is not None:
            person_data["spatialDescription"] = Resource._from_json_(sd, context)

        return cls(**person_data)   


from typing import Optional
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class SourceCitation:
    identifier = 'http://gedcomx.org/v1/SourceCitation'
    version = 'http://gedcomx.org/conceptual-model/v1'
    
    def __init__(self, lang: Optional[str], value: str) -> None:
        self.lang = lang if lang else 'en'
        self.value = value
    
    # ...existing code...

    @classmethod
    def _from_json_(cls, data: dict, context = None):
        """
        Create a SourceCitation instance from a JSON-dict (already parsed).
        """
        object_data = {}
        if (lang := data.get('lang')) is not None:
            object_data['lang'] = lang
        if (value := data.get('value')) is not None:
            object_data['value'] = value
        return cls(**object_data)
    
    @property
    def _as_dict_(self):
        return {'lang':self.lang,
                'value': self.value}


from enum import Enum
from typing import List, Optional
"""
======================================================================
 Project: Gedcom-X
 File:    gender.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactor 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .attribution import Attribution
from .conclusion import ConfidenceLevel, Conclusion
from .Extensions.rs10.rsLink import _rsLinks
from .note import Note
from .resource import Resource
from .source_reference import SourceReference
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class GenderType(Enum):
    Male = "http://gedcomx.org/Male"
    Female = "http://gedcomx.org/Female"
    Unknown = "http://gedcomx.org/Unknown"
    Intersex = "http://gedcomx.org/Intersex"
    
    @property
    def description(self):
        descriptions = {
            GenderType.Male: "Male gender.",
            GenderType.Female: "Female gender.",
            GenderType.Unknown: "Unknown gender.",
            GenderType.Intersex: "Intersex (assignment at birth)."
        }
        return descriptions.get(self, "No description available.")
    
class Gender(Conclusion):
    identifier = 'http://gedcomx.org/v1/Gender'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,
                 id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[Resource] = None,
                 notes: Optional[List[Note]] = None,
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None, 
                 type: Optional[GenderType] = None,
                 links: Optional[_rsLinks] = None
                 ) -> None:
        super().__init__(id=id, lang=lang, sources=sources, analysis=analysis, notes=notes, confidence=confidence, attribution=attribution, links=links)
        self.type = type
    
    @property
    def _as_dict_(self):
            
        type_as_dict = super()._as_dict_  or {}
        if self.type:
            type_as_dict['type'] = self.type.value if self.type else None                   
        
        return type_as_dict if type_as_dict != {} else None
        

    @classmethod
    def _from_json_(cls,data,context):
        conclusion = Conclusion._dict_from_json_(data,context)
        if (type_ := data.get("type")) is not None:
            conclusion["type"] = GenderType(type_)
        
        return cls(**conclusion)
        
        


from typing import Any, Optional, Dict

"""
======================================================================
 Project: Gedcom-X
 File:    textvalue.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03 _from_json_ refactor
   - 2025-09-04 added _str_ and _repr_ dunders
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class TextValue:
    identifier = 'http://gedcomx.org/v1/TextValue'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, value: Optional[str] = None, lang: Optional[str] = None) -> None:
        self.lang = lang
        self.value = value
    
    def _append_to_value(self, value_to_append):
        if not isinstance(value_to_append, str):
            raise ValueError(f"Cannot append object of type {type(value_to_append)}.")
        if self.value is None: self.value = value_to_append
        else: self.value += ' ' + value_to_append
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        return Serialization.serialize(self)
        type_as_dict = {}
        if self.lang is not None: type_as_dict['lang'] = self.lang
        if self.value is not None: type_as_dict['value'] = self.value
        return type_as_dict if type_as_dict != {} else None
    
    def __str__(self):
        return f"{self.value} ({self.lang})"
    
    def __eq__(self, other: object) -> bool:
        if isinstance(other,TextValue):
            if self.value.strip() == other.value.strip():
                return True
        return False
    
    # ...existing code...

    @classmethod
    def _from_json_(cls, data: Any, context: Any = None) -> "TextValue":
        """
        Create a TextValue from JSON.
        - Accepts a dict like {"value": "...", "lang": "en"} or a plain string shorthand.
        - Defaults lang to "en" if not provided.
        """
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)}")

        text_value_data: Dict[str, Any] = {}       
            
        if (v := data.get("value")) is None:
            text_value_data["value"] = v
        if (l := data.get("lang")) is None:
            text_value_data["lang"] = l

        return cls(**text_value_data)
    
    def __repr__(self) -> str:
        # Debug-friendly: unambiguous constructor-style representation
        cls = self.__class__.__name__
        return f"{cls}(value={self.value!r}, lang={self.lang!r})"



import warnings

from enum import Enum
from typing import Any, Dict, List, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from .document import Document


"""
======================================================================
 Project: Gedcom-X
 File:    SourceDescription.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-07-25
 Updated:
   - 2025-08-31: _as_dict_ refactored to ignore empty fields, changed id creation to make_uid()
    - 2025-09-01: filename PEP8 standard, imports changed accordingly
 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .agent import Agent
from .attribution import Attribution
from .coverage import Coverage
from .date import Date
from .identifier import Identifier, IdentifierList, make_uid
from .logging_hub import hub, logging
from .note import Note
from .resource import Resource
from .source_citation import SourceCitation
from .source_reference import SourceReference
from .textvalue import TextValue
from .uri import URI
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
deserial_log = "gedcomx.deserialization"
#=====================================================================


class ResourceType(Enum):
    Collection = "http://gedcomx.org/Collection"
    PhysicalArtifact = "http://gedcomx.org/PhysicalArtifact"
    DigitalArtifact = "http://gedcomx.org/DigitalArtifact"
    Record = "http://gedcomx.org/Record"
    Person = "http://gedcomx.org/Person"    
    
    @property
    def description(self):
        descriptions = {
            ResourceType.Collection: "A collection of genealogical resources. A collection may contain physical artifacts (such as a collection of books in a library), records (such as the 1940 U.S. Census), or digital artifacts (such as an online genealogical application).",
            ResourceType.PhysicalArtifact: "A physical artifact, such as a book.",
            ResourceType.DigitalArtifact: "A digital artifact, such as a digital image of a birth certificate or other record.",
            ResourceType.Record: "A historical record, such as a census record or a vital record."
        }
        return descriptions.get(self, "No description available.")
    
class SourceDescription:
    """Description of a genealogical information source.

    See: http://gedcomx.org/v1/SourceDescription

    Args:
        id (str | None): Unique identifier for this `SourceDescription`.
        resourceType (ResourceType | None): Type/category of the resource being
            described (e.g., digital artifact, physical artifact).
        citations (list[SourceCitation] | None): Citations that reference or
            justify this source description.
        mediaType (str | None): IANA media (MIME) type of the resource
            (e.g., ``"application/pdf"``).
        about (URI | None): Canonical URI that the description is about.
        mediator (Resource | None): The mediator resource (if any) involved in
            providing access to the source.
        publisher (Resource | Agent | None): Publisher of the resource.
        authors (list[Resource] | None): Authors/creators of the resource.
        sources (list[SourceReference] | None): Other sources this description
            derives from or references.
        analysis (Resource | None): Analysis document associated with the
            resource (often a `Document`; kept generic to avoid circular imports).
        componentOf (SourceReference | None): Reference to a parent/containing
            source (this is a component/child of that source).
        titles (list[TextValue] | None): One or more titles for the resource.
        notes (list[Note] | None): Human-authored notes about the resource.
        attribution (Attribution | None): Attribution metadata for who supplied
            or curated this description.
        rights (list[Resource] | None): Rights statements or licenses.
        coverage (list[Coverage] | None): Spatial/temporal coverage of the
            source’s content.
        descriptions (list[TextValue] | None): Short textual summaries or
            descriptions.
        identifiers (IdentifierList | None): Alternative identifiers for the
            resource (DOI, ARK, call numbers, etc.).
        created (Date | None): Creation date of the resource.
        modified (Date | None): Last modified date of the resource.
        published (Date | None): Publication/release date of the resource.
        repository (Agent | None): Repository or agency that holds the resource.
        max_note_count (int): Maximum number of notes to retain/emit. Defaults to 20.

    Raises:
        ValueError: If `id` is not a valid UUID.

    Attributes:
        identifier (str): Gedcom-X specification identifier for this type.
    """

    identifier = "http://gedcomx.org/v1/SourceDescription"
    version = 'http://gedcomx.org/conceptual-model/v1'
    
    def __init__(self, id: Optional[str] = None,
                 resourceType: Optional[ResourceType] = None,
                 citations: Optional[List[SourceCitation]] = [],
                 mediaType: Optional[str] = None,
                 about: Optional[URI] = None,
                 mediator: Optional[Agent|Resource] = None,
                 publisher: Optional[Agent|Resource] = None,
                 authors: Optional[List[Resource]] = None,
                 sources: Optional[List[SourceReference]] = None, # SourceReference
                 analysis: Optional["Document|Resource"] = None,  #TODO add type checker so its a document
                 componentOf: Optional[SourceReference] = None, # SourceReference
                 titles: Optional[List[TextValue]] = None,
                 notes: Optional[List[Note]] = None,
                 attribution: Optional[Attribution] = None,
                 rights: Optional[List[Resource]] = [],
                 coverage: Optional[List[Coverage]] = None, # Coverage
                 descriptions: Optional[List[TextValue]] = None,
                 identifiers: Optional[IdentifierList] = None,
                 created: Optional[Date] = None,
                 modified: Optional[Date] = None,
                 published: Optional[Date] = None,
                 repository: Optional[Agent] = None,
                 max_note_count: int = 20):
        
        self.id = id if id else make_uid()
        self.resourceType = resourceType
        self.citations = citations or []
        self.mediaType = mediaType
        self.about = about
        self.mediator = mediator
        self._publisher = publisher 
        self.authors = authors or []
        self.sources = sources or []
        self.analysis = analysis
        self.componentOf = componentOf
        self.titles = titles or []
        self.notes = notes or []
        self.attribution = attribution
        self.rights = rights or []
        self.coverage = coverage or []
        self.descriptions = descriptions or []
        self.identifiers = identifiers or IdentifierList()
        self.created = created
        self.modified = modified
        self.published = published
        self.repository = repository
        self.max_note_count = max_note_count

        self.uri = URI(fragment=id) if id else None #TODO Should i take care of this in the collections?

        self._place_holder = False
    
    @property
    def publisher(self) -> Resource | Agent | None:
        return self._publisher
    

    @publisher.setter
    def publisher(self, value: Resource | Agent):
        if value is None:
            self._publisher = None
        elif isinstance(value,Resource):
            self._publisher = value
        elif isinstance(value,Agent):
            self._publisher = value
        else:
            raise ValueError(f"'publisher' must be of type 'URI' or 'Agent', type: {type(value)} was provided")
    
    def add_description(self, desccription_to_add: TextValue):
        if desccription_to_add and isinstance(desccription_to_add,TextValue):
            for current_description in self.descriptions:
                if desccription_to_add == current_description:
                    return
            self.descriptions.append(desccription_to_add)

    def add_identifier(self, identifier_to_add: Identifier):
        if identifier_to_add and isinstance(identifier_to_add,Identifier):
            self.identifiers.append(identifier_to_add)
    
    def add_note(self,note_to_add: Note):
        if len(self.notes) >= self.max_note_count:
            warnings.warn(f"Max not count of {self.max_note_count} reached for id: {self.id}")
            return False
        if note_to_add and isinstance(note_to_add,Note):
            for existing in self.notes:
                if note_to_add == existing:
                    return False
            self.notes.append(note_to_add)
    
    def add_source_reference(self, source_to_add: SourceReference):
        if source_to_add and isinstance(object,SourceReference):
            for current_source in self.sources:
                if current_source == source_to_add:
                    return
            self.sources.append(source_to_add)

    def add_title(self, title_to_add: TextValue):
        if isinstance(title_to_add,str): title_to_add = TextValue(value=title_to_add)
        if title_to_add and isinstance(title_to_add, TextValue):
            for current_title in self.titles:
                if title_to_add == current_title:
                    return False
            self.titles.append(title_to_add)
        else:
            raise ValueError(f"Cannot add title of type {type(title_to_add)}")
            
    @property
    def _as_dict_(self) -> Dict[str, Any] | None:
        from .serialization import Serialization
        return Serialization.serialize(self)
        with hub.use(serial_log):
            log.debug(f"Serializing 'SourceDescription' with id: {self.id}")
            type_as_dict = {}

            if self.id:
                type_as_dict['id'] = self.id
            if self.about:
                type_as_dict['about'] = self.about._as_dict_
            if self.resourceType:
                if isinstance(self.resourceType,str):
                    log.warning(f"'SourceDescription.resourceType' should not be a string {self.resourceType}")
                    type_as_dict['resourceType'] = self.resourceType
                else:
                    type_as_dict['resourceType'] = self.resourceType.value 
            if self.citations:
                type_as_dict['citations'] = [c._as_dict_ for c in self.citations if c]
            if self.mediaType:
                type_as_dict['mediaType'] = self.mediaType
            
            if self.mediator:
                type_as_dict['mediator'] = self.mediator._as_dict_
            if self.publisher:
                type_as_dict['publisher'] = self.publisher._as_dict_ #TODO Resource this
            if self.authors:
                type_as_dict['authors'] = [a._as_dict_ for a in self.authors if a]
            if self.sources:
                type_as_dict['sources'] = [s._as_dict_ for s in self.sources if s]
            
            
            if self.analysis:
                type_as_dict['analysis'] = self.analysis._as_dict_
            if self.componentOf:
                type_as_dict['componentOf'] = self.componentOf._as_dict_ 
            if self.titles:
                type_as_dict['titles'] = [t._as_dict_ for t in self.titles if t]
            if self.notes:
                type_as_dict['notes'] = [n._as_dict_ for n in self.notes if n]
            if self.attribution:
                type_as_dict['attribution'] = self.attribution._as_dict_
            if self.rights:
                type_as_dict['rights'] = [r._as_dict_ for r in self.rights if r]
            
            if self.coverage:
                type_as_dict['coverage'] = [c._as_dict_ for c in self.coverage if c]
            
            if self.descriptions:
                if not (isinstance(self.descriptions, list) and all(isinstance(x, TextValue) for x in self.descriptions)):
                    assert False
                type_as_dict['descriptions'] = [d._as_dict_ for d in self.descriptions if d]
                
            if self.identifiers:
                type_as_dict['identifiers'] = self.identifiers._as_dict_
            
            if self.created is not None:
                type_as_dict['created'] = self.created
            if self.modified is not None:
                type_as_dict['modified'] = self.modified
            if self.published is not None:
                type_as_dict['published'] = self.published
            
            if self.repository:
                type_as_dict['repository'] = self.repository._as_dict_ #TODO Resource this       

            log.debug(f"'SourceDescription' serialized with fields: '{type_as_dict.keys()}'") 
            if type_as_dict == {}: log.warning("serializing and empty 'SourceDescription'")
        return type_as_dict if type_as_dict != {} else None
        
              
    @classmethod
    def _from_json_(cls, data: Dict[str, Any], context: Any = None) -> "SourceDescription":
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict, got {type(data)}")

        source_description_data: Dict[str, Any] = {}

        # ── Scalars ──────────────────────────────────────────────────────────────
        if (id_ := data.get("id")) is not None:
            source_description_data["id"] = id_
        if (media_type := data.get("mediaType")) is not None:
            source_description_data["mediaType"] = media_type
        if (mnc := data.get("max_note_count")) is not None:
            source_description_data["max_note_count"] = int(mnc)

        # resourceType (enum or similar)
        if (rt := data.get("resourceType")) is not None:
            source_description_data["resourceType"] = ResourceType(rt)

        # about: URI (accept string or dict)
        if (about := data.get("about")) is not None:
            source_description_data["about"] = URI.from_url(about) if isinstance(about, str) else URI._from_json_(about, context)

        # mediator / publisher can be Agent or Resource
        if (mediator := data.get("mediator")) is not None:
            if isinstance(mediator, dict) and any(k in mediator for k in ("names", "emails", "accounts", "addresses", "person")):
                source_description_data["mediator"] = Agent._from_json_(mediator, context)
            else:
                source_description_data["mediator"] = Resource._from_json_(mediator, context)

        if (publisher := data.get("publisher")) is not None:
            if isinstance(publisher, dict) and any(k in publisher for k in ("names", "emails", "accounts", "addresses", "person")):
                source_description_data["publisher"] = Agent._from_json_(publisher, context)
            else:
                source_description_data["publisher"] = Resource._from_json_(publisher, context)

        # repository (Agent)
        if (repo := data.get("repository")) is not None:
            source_description_data["repository"] = Agent._from_json_(repo, context)

        # analysis: Document | Resource   (prefer Document when possible)
        if (analysis := data.get("analysis")) is not None:
            try:
                source_description_data["analysis"] = Document._from_json_(analysis, context)
            except Exception:
                source_description_data["analysis"] = Resource._from_json_(analysis, context)

        # componentOf: SourceReference
        if (component := data.get("componentOf")) is not None:
            source_description_data["componentOf"] = SourceReference._from_json_(component, context)

        # identifiers
        if (identifiers := data.get("identifiers")) is not None:
            source_description_data["identifiers"] = IdentifierList._from_json_(identifiers, context)

        # timestamps
        if (created := data.get("created")) is not None:
            source_description_data["created"] = created 
        if (modified := data.get("modified")) is not None:
            source_description_data["modified"] = modified
        if (published := data.get("published")) is not None:
            source_description_data["published"] = published

        # ── Lists ───────────────────────────────────────────────────────────────
        if (citations := data.get("citations")) is not None:
            source_description_data["citations"] = [SourceCitation._from_json_(c, context) for c in citations]

        if (authors := data.get("authors")) is not None:
            source_description_data["authors"] = [Resource._from_json_(a, context) for a in authors]

        if (sources := data.get("sources")) is not None:
            source_description_data["sources"] = [SourceReference._from_json_(s, context) for s in sources]

        if (titles := data.get("titles")) is not None:
            source_description_data["titles"] = [
                (TextValue._from_json_(t, context) if isinstance(t, dict) else TextValue(t))
                for t in titles
            ]

        if (notes := data.get("notes")) is not None:
            source_description_data["notes"] = [Note._from_json_(n, context) for n in notes]

        if (rights := data.get("rights")) is not None:
            source_description_data["rights"] = [Resource._from_json_(r, context) for r in rights]

        if (coverage := data.get("coverage")) is not None:
            source_description_data["coverage"] = [Coverage._from_json_(c, context) for c in coverage]

        if (descriptions := data.get("descriptions")) is not None:
            for d in descriptions:
                if not isinstance(d, dict): assert False
            source_description_data["descriptions"] = [
                (TextValue._from_json_(d, context) if isinstance(d, dict) else TextValue(value="somethings fucked up"))
                for d in descriptions
            ]

        # attribution
        if (attr := data.get("attribution")) is not None:
            source_description_data["attribution"] = Attribution._from_json_(attr, context) if isinstance(attr, dict) else attr

        return cls(**source_description_data)

    


from __future__ import annotations

from typing import Dict, Any
import warnings
from . import specification as g7

from typing import Dict, List,Optional,Any


gedcom_top_level_terms = ['https://gedcom.io/terms/v7/CONT',
                          'https://gedcom.io/terms/v7/record-FAM',
                          'https://gedcom.io/terms/v7/record-INDI',
                          'https://gedcom.io/terms/v7/record-SNOTE',
                          'https://gedcom.io/terms/v7/record-SUBM',
                          'https://gedcom.io/terms/v7/TRLR',
                          'https://gedcom.io/terms/v7/HEAD',
                          'https://gedcom.io/terms/v7/record-OBJE',
                          'https://gedcom.io/terms/v7/record-REPO',
                          'https://gedcom.io/terms/v7/record-SOUR']


class GedcomStructure:
    version = 'v7'

    def __init__(
        self,
        *,
        level: int | None = None,
        xref: str | None = None,
        tag: str | None = None,
        pointer: bool | None = None,
        text: str | None = None,
        parent: GedcomStructure | None = None,
        line_num: int | None = None
    ) -> None:
        """Create a GEDCOM structure node.

        Args:
            level: GEDCOM line level (0..n).
            xref: Optional cross-reference id (e.g., '@I1@').
            tag: GEDCOM tag (e.g., 'INDI', 'NAME').
            pointer: True if this line is a pointer, False if not, None if unknown.
            text: Literal text payload for this line.
            parent: Parent node in the structure tree, if any.
        """
        self.level = level
        self.xref = xref
        self.tag = tag
        self.pointer = pointer
        self.text = text
        self.parent = parent
        self.value = text
        self.pointer = pointer if pointer else False
        self.line_num = line_num
        
        if self.level and self.level > 0 and text and text.startswith('@') and text.endswith('@'):
            self.pointer = True
            self.xref = text
        
        self.parent: GedcomStructure | None = parent if parent else None
        if self.parent and isinstance(self.parent, GedcomStructure):
            parent.subtructures.append(self)

        self.extension = False if not tag else True if tag.startswith('_') else False 
        self.uri = g7.match_uri(tag,self.parent) 
        self.label = g7.get_label(self.uri)
        
        self.subtructures = []
        

    def _as_dict_(self):
        as_dict =  {}
        as_dict['level'] = self.level
        if self.xref: as_dict['xref'] = self.xref
        as_dict['tag'] = self.tag
        if self.value: as_dict['value'] = self.value
        if self.subtructures: as_dict['substructures'] = [substructure._as_dict_() for substructure in self.subtructures]
        return {g7.get_label(self.uri):as_dict}
       
    def __repr__(self):
        return (
            "GedcomStructure("
            f"level: {self.level} tag={self.tag:<6} ({self.label}), {'(Ext)' if self.extension else ''} xref:{self.xref}  pointer={self.pointer}, text='{self.value}',  "
            f"uri={self.uri} subStructures: {len(self.subtructures)}"
        )
    
    def __getitem__(self,index) -> List['GedcomStructure']:
        return [s for s in self.subtructures if s.tag == index]

    
    
    